<!-- SPDX-License-Identifier: LGPL-2.1-or-later -->

<head>
  <meta charset="utf-8">
  <meta http-equiv="X-UA-Compatible" content="IE=edge">
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <meta name="theme-color" content="#201A26">

  <title>{% if page.title %}{{ page.title }}{% else %}{{ site.title }}{% endif %}</title>

  <link rel="canonical" href="{{ page.url | replace:'index.html','' | prepend: site.baseurl | prepend: site.url }}">

  <link rel="stylesheet" href="{{ "/style.css" | prepend: site.baseurl }}">

  <link rel="icon" type="image/png" href="/favicon.png" />
</head>
