/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

#include "sd-varlink-idl.h"

extern const sd_varlink_interface vl_interface_io_systemd_Machine;

extern const sd_varlink_symbol vl_type_MachineClass;
extern const sd_varlink_symbol vl_type_KillWhom;
