/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

#include "sd-varlink-idl.h"

extern const sd_varlink_interface vl_interface_io_systemd_Network;
extern const sd_varlink_symbol vl_type_LinkAddressState;
extern const sd_varlink_symbol vl_type_LinkOnlineState;
extern const sd_varlink_symbol vl_type_LinkRequiredAddressFamily;
extern const sd_varlink_symbol vl_type_LinkState;
extern const sd_varlink_symbol vl_type_Route;
extern const sd_varlink_symbol vl_type_NextHopGroup;
extern const sd_varlink_symbol vl_type_NextHop;
extern const sd_varlink_symbol vl_type_LLDPNeighbor;
extern const sd_varlink_symbol vl_type_DNS;
extern const sd_varlink_symbol vl_type_NTP;
extern const sd_varlink_symbol vl_type_SIP;
extern const sd_varlink_symbol vl_type_Domain;
extern const sd_varlink_symbol vl_type_DNSSECNegativeTrustAnchor;
extern const sd_varlink_symbol vl_type_DNSSetting;
extern const sd_varlink_symbol vl_type_Pref64;
extern const sd_varlink_symbol vl_type_NDisc;
extern const sd_varlink_symbol vl_type_Address;
extern const sd_varlink_symbol vl_type_Neighbor;
extern const sd_varlink_symbol vl_type_DHCPLease;
extern const sd_varlink_symbol vl_type_PrivateOption;
extern const sd_varlink_symbol vl_type_DHCPv6ClientPD;
extern const sd_varlink_symbol vl_type_DHCPv6ClientVendorOption;
extern const sd_varlink_symbol vl_type_DHCPv6Client;
extern const sd_varlink_symbol vl_type_DHCPServerLease;
extern const sd_varlink_symbol vl_type_DHCPServer;
extern const sd_varlink_symbol vl_type_BitRates;
extern const sd_varlink_symbol vl_type_Interface;
