/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

/* IWYU pragma: always_keep */

#include "conf-parser-forward.h"    /* IWYU pragma: export */
#include "shared-forward.h"         /* IWYU pragma: export */

typedef enum InhibitWhat InhibitWhat;
typedef enum KillWhom KillWhom;
typedef enum SessionType SessionType;
typedef enum SessionClass SessionClass;

typedef struct HandleActionData HandleActionData;
typedef struct Device Device;
typedef struct Manager Manager;
typedef struct Seat Seat;
typedef struct SessionDevice SessionDevice;
typedef struct Session Session;
typedef struct User User;
typedef struct Inhibitor Inhibitor;
typedef struct Button Button;
