/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

/* IWYU pragma: always_keep */

#include "shared-forward.h" /* IWYU pragma: export */

typedef struct UdevCtrl UdevCtrl;
typedef struct UdevEvent UdevEvent;
typedef struct Manager Manager;
typedef struct Event Event;
typedef struct UdevRules UdevRules;
typedef struct Worker Worker;
typedef struct UdevWorker UdevWorker;
typedef struct UdevRuleFile UdevRuleFile;
