/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

#include "sd-varlink-idl.h"

extern const sd_varlink_interface vl_interface_io_systemd_Unit;

extern const sd_varlink_symbol vl_type_ExecInputType;
extern const sd_varlink_symbol vl_type_ExecUtmpMode;
extern const sd_varlink_symbol vl_type_ExecPreserveMode;
extern const sd_varlink_symbol vl_type_ExecKeyringMode;
extern const sd_varlink_symbol vl_type_MemoryTHP;
extern const sd_varlink_symbol vl_type_ProtectProc;
extern const sd_varlink_symbol vl_type_ProcSubset;
extern const sd_varlink_symbol vl_type_ProtectSystem;
extern const sd_varlink_symbol vl_type_ProtectHome;
extern const sd_varlink_symbol vl_type_PrivateTmp;
extern const sd_varlink_symbol vl_type_PrivateUsers;
extern const sd_varlink_symbol vl_type_ProtectHostname;
extern const sd_varlink_symbol vl_type_ProtectControlGroups;
extern const sd_varlink_symbol vl_type_PrivatePIDs;
extern const sd_varlink_symbol vl_type_PrivateBPF;
extern const sd_varlink_symbol vl_type_CGroupDevicePolicy;
extern const sd_varlink_symbol vl_type_ManagedOOMPreference;
extern const sd_varlink_symbol vl_type_CGroupController;
extern const sd_varlink_symbol vl_type_CPUSchedulingPolicy;
extern const sd_varlink_symbol vl_type_IOSchedulingClass;
extern const sd_varlink_symbol vl_type_NUMAPolicy;
extern const sd_varlink_symbol vl_type_MountPropagationFlag;
extern const sd_varlink_symbol vl_type_KillMode;
extern const sd_varlink_symbol vl_type_AutomountResult;
extern const sd_varlink_symbol vl_type_MountResult;
extern const sd_varlink_symbol vl_type_PathType;
extern const sd_varlink_symbol vl_type_PathResult;
extern const sd_varlink_symbol vl_type_ScopeResult;
extern const sd_varlink_symbol vl_type_SocketBindIPv6Only;
extern const sd_varlink_symbol vl_type_SocketTimestamping;
extern const sd_varlink_symbol vl_type_SocketDeferTrigger;
extern const sd_varlink_symbol vl_type_SocketResult;
extern const sd_varlink_symbol vl_type_SwapResult;
extern const sd_varlink_symbol vl_type_TimerBase;
extern const sd_varlink_symbol vl_type_TimerResult;
extern const sd_varlink_symbol vl_type_CollectMode;
extern const sd_varlink_symbol vl_type_JobMode;
extern const sd_varlink_symbol vl_type_NotifyAccess;
extern const sd_varlink_symbol vl_type_ServiceExitType;
extern const sd_varlink_symbol vl_type_ServiceRestart;
extern const sd_varlink_symbol vl_type_ServiceRestartMode;
extern const sd_varlink_symbol vl_type_ServiceResult;
extern const sd_varlink_symbol vl_type_ServiceTimeoutFailureMode;
extern const sd_varlink_symbol vl_type_ServiceType;
