/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

#include "shared-forward.h"

int action_print_header(void);
int action_verify(void);
int action_disk_usage(void);
int action_list_boots(void);
int action_list_fields(void);
int action_list_field_names(void);
int action_list_invocations(void);
int action_list_namespaces(void);
