/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

#include "core-forward.h"
#include "crash-handler.h"

extern bool arg_dump_core;
extern int arg_crash_chvt;
extern bool arg_crash_shell;
extern CrashAction arg_crash_action;
