/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

/* IWYU pragma: always_keep */

#include "shared-forward.h" /* IWYU pragma: export */

typedef enum HomeSetupFlags HomeSetupFlags;

typedef struct PasswordCache PasswordCache;
typedef struct HomeSetup HomeSetup;
