/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

#include "shared-forward.h"

int verb_is_active(int argc, char *argv[], uintptr_t _data, void *userdata);
int verb_is_failed(int argc, char *argv[], uintptr_t _data, void *userdata);
