/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

void help_cmdline(const char *arguments);

void help_abstract(const char *text);

void help_section(const char *title);

void help_man_page_reference(const char *page, const char *section);
