/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

/* IWYU pragma: always_keep */

#include "shared-forward.h" /* IWYU pragma: export */

typedef struct Context Context;
typedef struct PartitionInfo PartitionInfo;
typedef struct Resource Resource;
typedef struct Instance Instance;
typedef struct InstanceMetadata InstanceMetadata;
