/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

/***
  Copyright © 2010 ProFUSION embedded systems
***/

#include "shared-forward.h"

int loopback_detach_all(bool *changed, bool last_try);
