/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

/* IWYU pragma: always_keep */

#include "conf-parser-forward.h"    /* IWYU pragma: export */
#include "shared-forward.h"         /* IWYU pragma: export */

typedef enum NetDevLocalAddressType NetDevLocalAddressType;
typedef enum NetworkConfigSource NetworkConfigSource;

typedef struct Address Address;
typedef struct IPv6Token IPv6Token;
typedef struct Link Link;
typedef struct Manager Manager;
typedef struct NetDev NetDev;
typedef struct Network Network;
typedef struct NextHop NextHop;
typedef struct Request Request;
typedef struct Route Route;
typedef struct Wireguard Wireguard;
